"""
Delta Lake source for Kukur

Two formats are supported:
- row based, with many series in one file containing one row per data point
- pivot, with many series as columns in one file
"""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0


try:
    from deltalake import DeltaTable

    HAS_DELTA_LAKE = True
except ImportError:
    HAS_DELTA_LAKE = False

from pyarrow import Table

from kukur.exceptions import InvalidSourceException, MissingModuleException
from kukur.source.arrow import BaseArrowSource
from kukur.source.quality import QualityMapper


class DeltaLakeLoader:
    """Fakes a loader for Delta Lake tables.

    It does not really load files, as the other loaders do."""

    def __init__(self, config: dict) -> None:
        self.__uri = config["uri"]

    def open(self):
        """Return the URI to connect to."""
        return self.__uri

    def has_child(self, subpath: str) -> bool:
        """Not supported for Delta Lake"""
        raise NotImplementedError()

    def open_child(self, subpath: str):
        """Not supported for Delta Lake"""
        raise NotImplementedError()


class DeltaLakeSource(BaseArrowSource):
    """Connect to a Delta Lake"""

    def read_file(self, file_like) -> Table:
        """Return a PyArrow Table for the Delta Table at the given URI."""
        return DeltaTable(file_like).to_pyarrow_table()

    def get_file_extension(self) -> str:
        """Delta lakes do not support row-based formats."""
        raise NotImplementedError()


def from_config(config: dict, quality_mapper: QualityMapper) -> DeltaLakeSource:
    """Create a new delta lake data source from the given configuration dictionary."""
    if not HAS_DELTA_LAKE:
        raise MissingModuleException("deltalake", "delta")

    data_format = config.get("format", "row")
    if data_format not in ["row", "pivot"]:
        raise InvalidSourceException(
            'Delta lake sources support only the "row" and "pivot" format.'
        )
    if "uri" not in config:
        raise InvalidSourceException('Delta lake sources require an "uri" entry')
    return DeltaLakeSource(
        data_format,
        DeltaLakeLoader(config),
        quality_mapper,
        sort_by_timestamp=config.get("sort_by_timestamp", False),
    )



"""The main objects in Kukur."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, field as data_field
from enum import Enum
from typing import Any, Dict, List, Optional, Union


@dataclass
class Dictionary:
    """A Dictionary maps numbers to labels.

    Time series can contain integer values that have a meaning. For example
    the number 0 could be 'OFF' and 1 'ON'.

    The ordering of the labels can also be significant while presenting the
    series to a user.

    In Python 3.8+, iteration over a dict keeps the insert ordering."""

    mapping: Dict[int, str]


@dataclass
class SeriesSearch:
    """SeriesSearch is the series selector to search series.

    The field can be optional to allow searching for anything.
    """

    source: str
    tags: Dict[str, str] = data_field(default_factory=dict)
    field: Optional[str] = None

    def __init__(
        self,
        source: str,
        tags: Union[str, Dict[str, str]] = None,
        field: Optional[str] = None,
    ):
        tags_dict = {}
        if isinstance(tags, str):
            tags_dict["series name"] = tags
        if isinstance(tags, dict):
            tags_dict = tags
        self.source = source
        self.tags = tags_dict
        self.field = field

    @property
    def name(self) -> str:
        """Get the series name with tags and fields included.

        For sources that cannot handle tags and fields yet."""
        series_tags: List[str] = []
        for tag_key, tag_value in self.tags.items():
            if tag_key == "series name":
                series_tags.insert(0, tag_value)
                continue
            series_tags.append(f"{tag_key}={tag_value}")
        series_string = ",".join(series_tags)
        if self.field is None:
            return f"{series_string}"
        return f"{series_string}::{self.field}"

    def to_data(self) -> Dict[str, Any]:
        """Convert to JSON object."""
        return dict(source=self.source, tags=self.tags, field=self.field)


@dataclass
class SeriesSelector(SeriesSearch):
    """SeriesSelector identifies a group of time series matching the given pattern."""

    field: str = "value"

    def __init__(
        self,
        source: str,
        tags: Union[str, Dict[str, str]] = None,
        field: str = "value",
    ):
        super().__init__(source, tags)
        self.field = field

    @classmethod
    def from_tags(cls, source: str, tags: Dict[str, str], field: Optional[str] = None):
        """Create the SeriesSelector from tags."""
        if field is None:
            field = "value"
        return cls(source, tags, field)

    @classmethod
    def from_data(cls, data: Dict[str, Any]):
        """Create a Series from a dictionary."""
        tags = data.get("tags", {})
        if "name" in data and "tags" not in data:
            tags["series name"] = data["name"]
        return cls(data["source"], tags, data.get("field", "value"))

    def to_data(self) -> Dict[str, Any]:
        """Convert to JSON object."""
        return dict(source=self.source, tags=self.tags, field=self.field)

    @property
    def name(self) -> str:
        """Get the series name with tags and fields included.

        For sources that cannot handle tags and fields yet."""
        series_tags: List[str] = []
        for tag_key, tag_value in self.tags.items():
            if tag_key == "series name":
                series_tags.insert(0, tag_value)
                continue
            series_tags.append(f"{tag_key}={tag_value}")
        series_string = ",".join(series_tags)
        if self.field == "value":
            return f"{series_string}"
        return f"{series_string}::{self.field}"


class InterpolationType(Enum):
    """InterpolationType describes how the value of a series evolves between data points."""

    LINEAR = "LINEAR"
    STEPPED = "STEPPED"


class DataType(Enum):
    """DataType represents the data type of the values in a time series.

    FLOAT32 is a 32 bit floating point number
    FLOAT64 is a 64 bit floating point number
    STRING is a variable length utf-8 character array
    DICTIONARY is an enumeration of ordered discrete values with a corresponding mapping
    CATEGORICAL is an enumeration of ordered discrete values
    """

    FLOAT32 = "FLOAT32"
    FLOAT64 = "FLOAT64"
    STRING = "STRING"
    DICTIONARY = "DICTIONARY"
    CATEGORICAL = "CATEGORICAL"


@dataclass
class SourceStructure:
    """The SourceStructure defines the fields and tags that are present in the source."""

    fields: List[str]
    tag_keys: List[str]
    tag_values: List[dict]

    @classmethod
    def from_data(cls, data: Dict[str, Any]) -> "SourceStructure":
        """Create a SourceStructure from a dictionary."""
        return cls(data["fields"], data["tagKeys"], data["tagValues"])

    def to_data(self):
        """Convert to JSON object."""
        return dict(
            fields=self.fields, tagKeys=self.tag_keys, tagValues=self.tag_values
        )



"""Connect to another Kukur instance using Arrow Flight."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Any, Dict, Generator, Optional, Tuple, Union

import pyarrow as pa

from kukur import Metadata, SeriesSearch, SeriesSelector, SourceStructure
from kukur.client import Client
from kukur.exceptions import InvalidDataError


def from_config(config: Dict[str, Any]):
    """Create a new Kukur source"""
    source_name = config["source"]
    host = config.get("host", "localhost")
    port = config.get("port", 8081)
    api_key = (config.get("api_key_name", ""), config.get("api_key", ""))
    return KukurSource(source_name, host, port, api_key)


class KukurSource:
    """KukurSource connects to another Kukur instance using Arrow Flight."""

    __client: Client
    __source_name: str

    # pylint: disable=too-many-arguments
    def __init__(
        self, source_name: str, host: str, port: int, api_key: Tuple[str, str]
    ):
        self.__client = Client(api_key, host, port)
        self.__source_name = source_name

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[Metadata, SeriesSelector], None, None]:
        """Search time series using the Flight service."""
        query = SeriesSearch(self.__source_name, selector.tags, selector.field)
        for result in self.__client.search(query):
            if isinstance(result, SeriesSelector):
                yield SeriesSelector.from_tags(
                    selector.source, result.tags, result.field
                )
            else:
                result.series = SeriesSelector.from_tags(
                    selector.source, result.series.tags, result.series.field
                )
                yield result

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Get metadata from the Flight service."""
        if "series name" not in selector.tags:
            raise InvalidDataError("No series name")
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        metadata = self.__client.get_metadata(remote_selector)
        metadata.series = SeriesSelector(selector.source, selector.tags, selector.field)
        return metadata

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Get data from the Flight service."""
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        return self.__client.get_data(remote_selector, start_date, end_date)

    def get_plot_data(
        self,
        selector: SeriesSelector,
        start_date: datetime,
        end_date: datetime,
        interval_count: int,
    ) -> pa.Table:
        """Get plot data from the Flight service."""
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        return self.__client.get_plot_data(
            remote_selector, start_date, end_date, interval_count
        )

    def get_source_structure(
        self, selector: SeriesSelector
    ) -> Optional[SourceStructure]:
        """Return the source structure using the Flight service."""
        remote_selector = SeriesSelector.from_tags(
            self.__source_name, selector.tags, selector.field
        )
        return self.__client.get_source_structure(remote_selector)



"""Optional support for Delta Lake using https://github.com/delta-io/delta-rs"""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from .delta_lake import from_config

__all__ = ["from_config"]



"""Connections to CrateDB data sources from Kukur.

This requires the crate package
"""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

try:
    from crate import client

    HAS_CRATE = True
except ImportError:
    HAS_CRATE = False

from kukur.exceptions import MissingModuleException
from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper
from kukur.source.sql import BaseSQLSource, SQLConfig


def from_config(
    data, metadata_value_mapper: MetadataValueMapper, quality_mapper: QualityMapper
):
    """Create a new CrateDB data source from a configuration dict.

    Raises ADODBNotInstalledError when the adodbapi module is not available."""
    if not HAS_CRATE:
        raise MissingModuleException("crate", "cratedb")

    config = SQLConfig.from_dict(data)

    return CrateDBSource(config, metadata_value_mapper, quality_mapper)


class CrateDBSource(BaseSQLSource):
    """A CrateDB data source."""

    def __init__(
        self,
        config: SQLConfig,
        metadata_value_mapper: MetadataValueMapper,
        quality_mapper: QualityMapper,
    ):
        super().__init__(config, metadata_value_mapper, quality_mapper)
        if not HAS_CRATE:
            raise MissingModuleException("crate", "adodb")

    def connect(self):
        return client.connect(self._config.connection_string)



"""Time series metadata.

Metadata in Kukur is flexible. Some fields are common to all time series.
They have been defined here. Users of Kukur can define their own metadata fields.
"""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, Generator, List, Optional, Tuple, TypeVar, Union

from kukur.base import SeriesSelector

from .fields import MetadataField, register_default_fields

T = TypeVar("T")


class MetadataFields:
    """Metadata fields is a collection of custom metadata fields and their values.

    Fields can be typed and registered to MetadataFields by providing a MetadataField object.
    Alternatively, untyped methods that accept fields by name are also available.
    """

    _fields: List[MetadataField] = []
    __values: Dict[Union[str, MetadataField], Any]

    def __init__(
        self,
        fields: List[MetadataField],
        values: Optional[Dict[MetadataField, Any]] = None,
    ):
        self._fields = fields
        self.__values = {k: k.default() for k in self._fields}
        if values is not None:
            for field, value in values.items():
                self.__values[field] = value

    def iter_fields(self) -> Generator[Tuple[MetadataField, Any], None, None]:
        """Iterate over all typed metadata fields and their values."""
        for field in self._fields:
            yield (field, field.calculated_value(self, self.__values.get(field)))

    def iter_names(self) -> Generator[Tuple[str, Any], None, None]:
        """Iterate over all metadata fields (typed and untyped) and return their names and values."""
        for field, value in self.__values.items():
            if isinstance(field, MetadataField):
                yield (field.name(), field.calculated_value(self, value))
            else:
                yield (field, value)

    def iter_serialized(self) -> Generator[Tuple[str, Any], None, None]:
        """Iterate over all metadata fields, but use the human readable names and serialized values."""
        for field, value in self.__values.items():
            if isinstance(field, MetadataField):
                yield (
                    field.name(),
                    field.serialize(field.calculated_value(self, value)),
                )
            else:
                yield (field, value)

    def find_field(self, field_name: str) -> MetadataField:
        """Return the MetadataField with the given name.

        Name can either be the human readable name or a camelCase name for JSON."""
        field = self._find_field(field_name)
        if field is None:
            raise AttributeError()
        return field

    def set_field(self, field: MetadataField[T], value: T):
        """Set the given field to the specified value."""
        if field not in self._fields:
            raise AttributeError()
        self.__values[field] = value

    def set_field_by_name(self, field_name: str, value: Any):
        """Set the field of the given name to the corresponding value."""
        field = self._find_field(field_name)
        if field is None:
            self.__values[field_name] = value
        else:
            self.__values[field] = value

    def coerce_field(self, field_name: str, value: Any):
        """Set the field of the given name to the corresponding value.

        This tries to coerce the given value into a correctly typed one."""
        field = self._find_field(field_name)
        if field is not None:
            self.__values[field] = field.deserialize(value)
        else:
            self.__values[field_name] = value

    def get_field(self, field: MetadataField[T]) -> T:
        """Return the value of the given field."""
        if field not in self._fields:
            raise AttributeError()
        return field.calculated_value(self, self.__values[field])

    def get_field_by_name(self, field_name: str) -> Any:
        """Return the value of the given field.

        When a field has not been defined, the value will be the serialized value."""
        field = self._find_field(field_name)
        if field is None:
            if field_name not in self.__values:
                return None
            return self.__values[field_name]
        return self.get_field(field)

    def to_data(self) -> Dict[str, Any]:
        """Convert the metadata to a Dictionary with camelcase keys as expected in JSON."""
        data = {}
        for field, value in self.__values.items():
            if isinstance(field, MetadataField):
                data[field.serialized_name()] = field.serialize(value)
            else:
                data[field] = value
        return data

    def _find_field(self, field_name: str) -> Optional[MetadataField]:
        for field in self._fields:
            if field.name() == field_name or field.serialized_name() == field_name:
                return field
        return None


class Metadata(MetadataFields):
    """Metadata fields for one time series."""

    _fields: List[MetadataField] = []
    series: SeriesSelector

    @classmethod
    def register_field(
        cls, field: MetadataField[T], *, after_field: Optional[MetadataField] = None
    ):
        """Register a new metadata field.

        Optionally insert it right after the given field in the field ordering."""
        if after_field is not None:
            cls._fields.insert(cls._fields.index(after_field) + 1, field)
        else:
            cls._fields.append(field)

    @classmethod
    def from_data(
        cls, data: Dict[str, Any], series: Optional[SeriesSelector] = None
    ) -> "Metadata":
        """Create a new Metadata object from a dictionary produced by to_data().

        This used the provided series as selector, otherwise the series is expected to be inside the data."""
        if series is None:
            if "series" not in data:
                raise AttributeError()
            series = SeriesSelector.from_data(data["series"])
        metadata = cls(series)
        for k, v in data.items():
            if k == "series":
                continue
            metadata.coerce_field(k, v)
        return metadata

    def __init__(
        self,
        series: SeriesSelector,
        values: Optional[Dict[MetadataField, Any]] = None,
    ):
        super().__init__(self._fields, values)
        self.series = series

    def __repr__(self) -> str:
        data = dict(self.iter_names())
        return f"Metadata({self.series}, {data})"

    def to_data(self) -> Dict[str, Any]:
        """Convert the metadata to a Dictionary with camelcase keys as expected in JSON."""
        data = super().to_data()
        data["series"] = self.series.to_data()
        return data


register_default_fields(Metadata)



"""Provide remote access to Kukur over Arrow Flight."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import json

from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import pyarrow as pa
import pyarrow.flight as fl

from kukur import Metadata, SeriesSearch, SeriesSelector, SourceStructure


class Client:
    """Client connects to Kukur using Arrow Flight."""

    _client: fl.FlightClient = None

    def __init__(
        self,
        api_key: Tuple[str, str] = ("", ""),
        host: str = "localhost",
        port: int = 8081,
    ):
        """Create a new Client.

        Creating a client does not open a connection. The connection will be opened lazily.

        Args:
            host: the hostname where the Kukur instance is running. Defaults to ``localhost``.
            port: the port where the Kukur instance is running. Defaults to ``8081``.
            api_key: the api key to connect: this is a tuple of (key name, key).
        """
        self._location = (host, port)
        self._api_key = api_key

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[Metadata, SeriesSelector], None, None]:
        """Search Kukur for time series matching the given ``SeriesSelector``.

        Args:
            selector: return time series matching the given selector.
                      Use ``name = None`` (the default) to select all series in a source.

        Returns:
            A generator that returns either ``Metadata`` or ``SeriesSelector``s.
            The return value depends on the search that is supported by the source.
        """
        body = selector.to_data()
        results = list(
            self._get_client().do_action(("search", json.dumps(body).encode()))
        )
        for result in results:
            data = json.loads(result.body.to_pybytes())
            if "series" not in data:
                yield SeriesSelector.from_data(data)
            else:
                yield _read_metadata(data)

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Read metadata for the time series selected by the ``SeriesSelector``.

        Args:
            selector: the selected time series

        Returns:
            The ``Metadata`` for the time series.
        """
        body = selector.to_data()
        results = list(
            self._get_client().do_action(("get_metadata", json.dumps(body).encode()))
        )
        result = results[0]
        data = json.loads(result.body.to_pybytes())
        return _read_metadata(data)

    def get_data(
        self,
        selector: SeriesSelector,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> pa.Table:
        """Get raw data for the time series selected by the SeriesSelector.

        Args:
            selector: return data for the time series selected by this selector.
            start_date: the start date of the time range of data to return. Defaults to one year ago.
            end_date: the end date of the time range of data to return. Defaults to now.

        Returns:
            A pyarrow Table with two columns: 'ts' and 'value'.
        """
        start_date, end_date = _apply_default_range(start_date, end_date)
        query = {
            "query": "get_data",
            "selector": selector.to_data(),
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
        }
        ticket = fl.Ticket(json.dumps(query))
        return self._get_client().do_get(ticket).read_all()

    def get_plot_data(
        self,
        selector: SeriesSelector,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        interval_count: int = 200,
    ) -> pa.Table:
        """Get plot data for the time series selected by the SeriesSelector.

        Args:
            selector: return data for the time series selected by this selector.
            start_date: the start date of the time range of data to return. Defaults to one year ago.
            end_date: the end date of the time range of data to return. Defaults to now.
            interval_count: the number of intervals included in the plot. Defaults to 200.

        Returns:
            A pyarrow Table with two columns: 'ts' and 'value'.
        """
        start_date, end_date = _apply_default_range(start_date, end_date)
        query = {
            "query": "get_plot_data",
            "selector": selector.to_data(),
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
            "interval_count": interval_count,
        }
        ticket = fl.Ticket(json.dumps(query))
        return self._get_client().do_get(ticket).read_all()

    def list_sources(self) -> List[str]:
        """List all configured sources.

        Returns:
            A list of source names that are configured in Kukur.
        """
        results = list(self._get_client().do_action(("list_sources")))
        data = json.loads(results[0].body.to_pybytes())
        return data

    def get_source_structure(
        self, selector: SeriesSelector
    ) -> Optional[SourceStructure]:
        """List all tags and fields from a source.

        Returns:
            A list of tag keys, tag values and fields that are configured in the source.
        """
        body = selector.to_data()
        results = list(
            self._get_client().do_action(
                ("get_source_structure", json.dumps(body).encode())
            )
        )
        data = json.loads(results[0].body.to_pybytes())
        if data is None:
            return None
        return SourceStructure.from_data(data)

    def _get_client(self):
        if self._client is None:
            self._client = fl.FlightClient(self._location)
            if self._api_key != ("", ""):
                self._client.authenticate(ClientAuthenticationHandler(self._api_key))
        return self._client


def _read_metadata(data: Dict[str, Any]) -> Metadata:
    return Metadata.from_data(data)


def _apply_default_range(
    start_date: Optional[datetime], end_date: Optional[datetime]
) -> Tuple[datetime, datetime]:
    if start_date is None or end_date is None:
        now = datetime.utcnow().replace(tzinfo=timezone(timedelta(0)))
        if start_date is None:
            start_date = now.replace(year=now.year - 1)
        if end_date is None:
            end_date = now
    return start_date, end_date


class ClientAuthenticationHandler(fl.ClientAuthHandler):
    """Client authentication handler for api keys"""

    def __init__(self, api_key):
        super().__init__()
        self.basic_auth = fl.BasicAuth(*api_key)
        self.token = None

    def authenticate(self, outgoing, incoming):
        """Client - server handshake"""
        auth = self.basic_auth.serialize()
        outgoing.write(auth)
        self.token = incoming.read()

    def get_token(self):
        """Get the token"""
        return self.token



"""Data sources for Kukur."""

import functools
import inspect
import logging
import time

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union

import pyarrow as pa
import pyarrow.types

from kukur.source import (
    adodb,
    cratedb,
    csv,
    azure_data_explorer,
    delta,
    feather,
    influxdb,
    kukur as kukur_source,
    odbc,
    parquet,
    piwebapi_da,
)

from kukur import (
    Metadata,
    PlotSource,
    SeriesSearch,
    SeriesSelector,
    Source as SourceProtocol,
    SourceStructure,
    TagSource,
)
from kukur.exceptions import InvalidSourceException
from kukur.source.quality import QualityMapper

from .metadata import MetadataMapper, MetadataValueMapper

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

logger = logging.getLogger(__name__)

_FACTORY = {
    "adodb": adodb.from_config,
    "cratedb": cratedb.from_config,
    "csv": csv.from_config,
    "azure-data-explorer": azure_data_explorer.from_config,
    "delta": delta.from_config,
    "feather": feather.from_config,
    "influxdb": influxdb.from_config,
    "kukur": kukur_source.from_config,
    "odbc": odbc.from_config,
    "parquet": parquet.from_config,
    "piwebapi-da": piwebapi_da.from_config,
}


@dataclass
class MetadataSource:
    """A metadata source provides at least some metadata fields."""

    source: SourceProtocol
    fields: List[str] = field(default_factory=list)


@dataclass
class Source:
    """A Kukur source can contain different metadata and data sources.

    Source keeps them together."""

    metadata: Union[SourceProtocol, TagSource]
    data: SourceProtocol


def _retry(retry_count: int, retry_delay: float, data_fn, log_message: str):
    while True:
        try:
            return data_fn()
        except Exception as err:  # pylint: disable=broad-except
            retry_count = retry_count - 1
            if retry_count < 0:
                raise err
            logger.info(
                "%s, %d retries left",
                log_message,
                retry_count + 1,
                exc_info=True,
            )
            time.sleep(retry_delay)


class SourceWrapper:
    """Handles query policy for a data source.

    It ensures requests do not overload a source.

    Metadata for a source can be fetched from multiple external sources and will be merged here.
    """

    __source: Source
    __metadata: List[MetadataSource]
    __query_retry_count: int
    __query_retry_delay: float
    __data_query_interval: Optional[timedelta] = None

    def __init__(
        self,
        source: Source,
        metadata_sources: List[MetadataSource],
        common_options,
    ):
        self.__source = source
        self.__metadata = metadata_sources
        self.__query_retry_count = common_options.get("query_retry_count", 0)
        self.__query_retry_delay = common_options.get("query_retry_delay", 1.0)
        if "data_query_interval_seconds" in common_options:
            self.__data_query_interval = timedelta(
                seconds=common_options["data_query_interval_seconds"]
            )

    def search(
        self, selector: SeriesSearch
    ) -> Generator[Union[SeriesSelector, Metadata], None, None]:
        """Search for all time series matching the given selector.

        The result is either a sequence of selectors for each time series in the source or a sequence of metadata
        entries for all series in the source if fetching the metadata can be done in the same request.

        If metadata sources are configured, query them as well and merge the results. This means that sources that
        are fast to search because they return metadata now also result in one additional query to each metadata
        source for each series."""
        query_fn = functools.partial(self.__source.metadata.search, selector)
        results = _retry(
            self.__query_retry_count,
            self.__query_retry_delay,
            query_fn,
            f"Search query for {selector.source} failed",
        )
        if results is None:
            return
        for result in results:
            if (
                len(self.__metadata) == 0
                or isinstance(result, SeriesSelector)
                or "series name" not in result.series.tags
            ):
                yield result
            else:
                extra_metadata = self.get_metadata(
                    SeriesSelector.from_tags(
                        result.series.source, result.series.tags, result.series.field
                    )
                )
                for k, v in result.iter_names():
                    if v is not None and v != "":
                        extra_metadata.set_field_by_name(k, v)
                yield extra_metadata

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Return the metadata for the given series.

        The resulting metadata is the combination of the metadata in the source itself and any additional
        metadata sources. Metadata sources earlier in the list of sources take precendence over later ones."""
        metadata = Metadata(selector)
        if "series name" not in selector.tags:
            return metadata
        for metadata_source in list(reversed(self.__metadata)) + [
            MetadataSource(self.__source.metadata)
        ]:
            query_fn = functools.partial(metadata_source.source.get_metadata, selector)
            received_metadata = _retry(
                self.__query_retry_count,
                self.__query_retry_delay,
                query_fn,
                f'Metadata query for "{selector.name}" ({selector.source}) failed',
            )
            if len(metadata_source.fields) == 0:
                for k, v in received_metadata.iter_names():
                    if v is not None and v != "":
                        metadata.set_field_by_name(k, v)
            else:
                for field_name in metadata_source.fields:
                    field_value = received_metadata.get_field_by_name(field_name)
                    if field_value is not None and field_value != "":
                        metadata.set_field_by_name(field_name, field_value)
        return metadata

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return the data for the given series in the given time frame, taking into account the request policy."""
        if start_date == end_date:
            return pa.Table.from_pydict({"ts": [], "value": [], "quality": []})
        tables = [
            self._get_data_chunk(selector, start, end)
            for start, end in self.__to_intervals(start_date, end_date)
        ]
        return concat_tables(tables)

    def get_plot_data(
        self,
        selector: SeriesSelector,
        start_date: datetime,
        end_date: datetime,
        interval_count: int,
    ) -> pa.Table:
        """Return the plotting data for the given series in the given time frame

        Plot data calls are not subject to request splitting, but do obsever other settings.

        Returns normal data when plot data is not supported."""
        if start_date == end_date:
            return pa.Table.from_pydict({"ts": [], "value": [], "quality": []})
        if not isinstance(self.__source.data, PlotSource):
            return self.get_data(selector, start_date, end_date)
        query_fn = functools.partial(
            self.__source.data.get_plot_data,
            selector,
            start_date,
            end_date,
            interval_count,
        )
        return _retry(
            self.__query_retry_count,
            self.__query_retry_delay,
            query_fn,
            f'Plot data query for "{selector.name}" ({selector.source}) ({start_date} to {end_date}) failed',
        )

    def get_source_structure(
        self, selector: SeriesSelector
    ) -> Optional[SourceStructure]:
        """Return the structure of the source for the given series."""
        if not isinstance(self.__source.metadata, TagSource):
            return None
        query_fn = functools.partial(
            self.__source.metadata.get_source_structure, selector
        )
        return _retry(
            self.__query_retry_count,
            self.__query_retry_delay,
            query_fn,
            f"Source structure query for {selector.source} failed",
        )

    def _get_data_chunk(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ):
        query_fn = functools.partial(
            self.__source.data.get_data, selector, start_date, end_date
        )
        return _retry(
            self.__query_retry_count,
            self.__query_retry_delay,
            query_fn,
            f'Data query for "{selector.name}" ({selector.source}) ({start_date} to {end_date}) failed',
        )

    def __to_intervals(
        self, start_date: datetime, end_date: datetime
    ) -> Generator[Tuple[datetime, datetime], None, None]:
        if self.__data_query_interval is None:
            yield (start_date, end_date)
            return
        current_date = start_date
        while current_date < end_date:
            next_date = current_date + self.__data_query_interval
            next_date = min(next_date, end_date)
            yield (current_date, next_date)
            current_date = next_date


class SourceFactory:
    """Source factory to create Source objects"""

    __config: Dict[str, Any]
    __factory: Dict[str, Callable]

    def __init__(self, config):
        self.__config = config
        self.__factory = _FACTORY.copy()

    def register_source(
        self, source_type_name: str, source_factory: Callable[..., SourceProtocol]
    ):
        """Register a new source type with the factory."""
        self.__factory[source_type_name] = source_factory

    def get_source_names(self) -> List[str]:
        """Get the data sources names as configured in the Kukur configuration."""
        sources = []
        for name, _ in self.__config.get("source", {}).items():
            sources.append(name)
        return sources

    def get_source(self, source_name: str) -> Optional[SourceWrapper]:
        """Get the data source and type as configured in the Kukur configuration."""
        metadata_sources = self._get_extra_metadata_sources()

        for name, options in self.__config.get("source", {}).items():
            if source_name == name:
                if "type" not in options:
                    raise InvalidSourceException(f'"{name}" has no type')
                source_type = options["type"]
                if source_type not in self.__factory:
                    raise InvalidSourceException(
                        f'Source "{name}" has unknown type "{source_type}"'
                    )
                data_source = self._make_source(source_type, options)
                metadata_source = data_source
                metadata_source_type = options.get("metadata_type", source_type)
                if metadata_source_type != source_type:
                    if metadata_source_type not in self.__factory:
                        raise InvalidSourceException(
                            f'"Source {name}" has unknown metadata type "{metadata_source_type}"'
                        )
                    metadata_source = self._make_source(metadata_source_type, options)

                extra_metadata = []
                for metadata_source_name in options.get("metadata_sources", []):
                    if metadata_source_name not in metadata_sources:
                        raise InvalidSourceException(
                            f'Metadata source "{metadata_source_name}" for source "{name}" not found'
                        )
                    extra_metadata.append(metadata_sources[metadata_source_name])

                return SourceWrapper(
                    Source(metadata_source, data_source), extra_metadata, options
                )
        return None

    def _get_extra_metadata_sources(self) -> Dict[str, MetadataSource]:
        metadata_sources = {}
        for name, options in self.__config.get("metadata", {}).items():
            if "type" not in options:
                raise InvalidSourceException(f'Metadata source "{name}" has no type.')
            source_type = options["type"]
            if source_type not in self.__factory:
                raise InvalidSourceException(
                    f'Metadata source "{name}" has unknown type "{source_type}"'
                )
            metadata_sources[name] = MetadataSource(
                self._make_source(source_type, options), options.get("fields", [])
            )
        return metadata_sources

    def _make_source(
        self, source_type: str, source_config: Dict[str, Any]
    ) -> Union[SourceProtocol, TagSource]:
        metadata_mapper = self._get_metadata_mapper(
            source_config.get("metadata_mapping")
        )
        metadata_value_mapper = self._get_metadata_value_mapper(
            source_config.get("metadata_value_mapping")
        )
        quality_mapper = self._get_quality_mapper(source_config.get("quality_mapping"))

        factory_function: Any = self.__factory[source_type]

        arguments: List[Any] = []
        for parameter in inspect.signature(factory_function).parameters.values():
            if parameter.annotation == MetadataMapper:
                arguments.append(metadata_mapper)
            elif parameter.annotation == MetadataValueMapper:
                arguments.append(metadata_value_mapper)
            elif parameter.annotation == QualityMapper:
                arguments.append(quality_mapper)
            else:
                arguments.append(source_config)

        return factory_function(*arguments)

    def _get_metadata_mapper(self, name: Optional[str]) -> MetadataMapper:
        if name is None or name not in self.__config.get("metadata_mapping", {}):
            return MetadataMapper()

        return MetadataMapper.from_config(self.__config["metadata_mapping"][name])

    def _get_metadata_value_mapper(self, name: Optional[str]) -> MetadataValueMapper:
        if name is None or name not in self.__config.get("metadata_value_mapping", {}):
            return MetadataValueMapper()

        return MetadataValueMapper.from_config(
            self.__config["metadata_value_mapping"][name]
        )

    def _get_quality_mapper(self, name: Optional[str]) -> QualityMapper:
        if name is None or name not in self.__config.get("quality_mapping", {}):
            return QualityMapper()

        return QualityMapper.from_config(self.__config["quality_mapping"][name])


def concat_tables(tables: List[pa.Table]) -> List[pa.Table]:
    """Safely concatenate multiple pyarrow.Table's.

    If any of the given tables contains strings, the result will contain a
    string value. If any of the given tables contains a floating point number,
    the value will be a double."""
    tables = [table for table in tables if len(table) > 0]
    if len(tables) == 0:
        return pa.Table.from_pydict({"ts": [], "value": [], "quality": []})

    schema = pa.schema(
        [
            ("ts", pa.timestamp("us", "utc")),
            ("value", pa.float64()),
        ]
    )

    if _has_any_string(tables):
        schema = pa.schema(
            [
                ("ts", pa.timestamp("us", "utc")),
                ("value", pa.string()),
            ]
        )

    if _is_all_integer(tables):
        schema = pa.schema(
            [
                ("ts", pa.timestamp("us", "utc")),
                ("value", pa.int64()),
            ]
        )

    if _has_quality_data_flag(tables):
        schema = schema.append(pa.field("quality", pa.int8()))

    return pa.concat_tables([table.cast(schema) for table in tables])


def _has_any_string(tables: List[pa.Table]) -> bool:
    string_tables = [
        table
        for table in tables
        if pyarrow.types.is_string(table.schema.field("value").type)
    ]
    return len(string_tables) > 0


def _is_all_integer(tables: List[pa.Table]) -> bool:
    integer_tables = [
        table
        for table in tables
        if pyarrow.types.is_integer(table.schema.field("value").type)
    ]
    return len(integer_tables) == len(tables)


def _has_quality_data_flag(tables: List[pa.Table]) -> bool:
    quality_table = [table for table in tables if "quality" in table.column_names]
    return len(quality_table) > 0



"""Time series metadata fields common to all time series."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Callable, Generic, List, Optional, Tuple, TypeVar

from kukur.base import (
    DataType as KukurDataType,
    Dictionary as KukurDictionary,
    InterpolationType as KukurInterpolationType,
)

T = TypeVar("T")


class MetadataField(Generic[T]):
    """A typed metadata field.

    This class is generic over the actual type of the field value.

    JSON-ready data dictionary conversions should be provided for all non-trivial implementations."""

    def __init__(
        self,
        name: str,
        *,
        default: T,
        serialized_name: str,
        serialize: Optional[Callable[[T], Any]] = None,
        deserialize: Optional[Callable[[Any], T]] = None,
        calculate: Optional[Callable[[Any, Any], T]] = None,
    ):
        self.__name = name
        self.__default = default
        self.__serialized_name = serialized_name
        self.__serialize = serialize
        self.__deserialize = deserialize
        self.__calculate = calculate

    def __repr__(self) -> str:
        return f'MetadataField("{self.__name}")'

    def __hash__(self):
        return hash(self.__name)

    def name(self) -> str:
        """Return the name of the metadata field."""
        return self.__name

    def default(self) -> T:
        """Return the default (empty) value for this field."""
        return self.__default

    def serialized_name(self) -> str:
        """Return the name of the field in a JSON dictionary."""
        return self.__serialized_name

    def serialize(self, value: T) -> Any:
        """Convert a field of this type to JSON."""
        if self.__serialize is None:
            return value
        return self.__serialize(value)

    def deserialize(self, value: Any) -> T:
        """Convert the result of a serialization to a proper value."""
        if self.__deserialize is None:
            return value
        return self.__deserialize(value)

    def calculated_value(self, metadata, value: Any) -> T:
        """Convert the value based on the calculate function."""
        if self.__calculate is None:
            return value
        return self.__calculate(metadata, value)


Description = MetadataField[str](
    "description", default="", serialized_name="description"
)

Unit = MetadataField[str]("unit", default="", serialized_name="unit")


def _parse_float(number: Optional[Any]) -> Optional[float]:
    if number is None:
        return None
    return float(number)


LimitLowPhysical = MetadataField[Optional[float]](
    "physical lower limit",
    default=None,
    serialized_name="limitLowPhysical",
    deserialize=_parse_float,
)


LimitHighPhysical = MetadataField[Optional[float]](
    "physical upper limit",
    default=None,
    serialized_name="limitHighPhysical",
    deserialize=_parse_float,
)


LimitLowFunctional = MetadataField[Optional[float]](
    "functional lower limit",
    default=None,
    serialized_name="limitLowFunctional",
    deserialize=_parse_float,
)


LimitHighFunctional = MetadataField[Optional[float]](
    "functional upper limit",
    default=None,
    serialized_name="limitHighFunctional",
    deserialize=_parse_float,
)


def _calculate_accuracy(metadata, accuracy: Optional[float]) -> Optional[float]:
    """Calculate the accuracy based on the accuracy percentage."""
    if accuracy is not None:
        return accuracy
    accuracy_percentage = metadata.get_field(AccuracyPercentage)
    if (
        accuracy_percentage is None
        or accuracy_percentage < 0
        or accuracy_percentage > 100
    ):
        return None
    low_limit = metadata.get_field(LimitLowPhysical)
    if low_limit is None:
        low_limit = metadata.get_field(LimitLowFunctional)
        if low_limit is None:
            return None
    high_limit = metadata.get_field(LimitHighPhysical)
    if high_limit is None:
        high_limit = metadata.get_field(LimitHighFunctional)
        if high_limit is None:
            return None
    return (high_limit - low_limit) * float(accuracy_percentage) / 100


Accuracy = MetadataField[Optional[float]](
    "accuracy",
    default=None,
    serialized_name="accuracy",
    deserialize=_parse_float,
    calculate=_calculate_accuracy,
)


def _parse_accuracy_percentage_float(number: Optional[Any]) -> Optional[float]:
    if number is None:
        return None
    return float(number)


AccuracyPercentage = MetadataField[Optional[float]](
    "accuracy percentage",
    default=None,
    serialized_name="accuracyPercentage",
    deserialize=_parse_accuracy_percentage_float,
)


def _interpolation_type_to_json(
    interpolation_type: Optional[KukurInterpolationType],
) -> Optional[str]:
    if interpolation_type is None:
        return None
    return interpolation_type.value


def _interpolation_type_from_json(
    interpolation_type: Optional[str],
) -> Optional[KukurInterpolationType]:
    if interpolation_type is None:
        return None
    return KukurInterpolationType(interpolation_type)


InterpolationType = MetadataField[Optional[KukurInterpolationType]](
    "interpolation type",
    default=None,
    serialized_name="interpolationType",
    serialize=_interpolation_type_to_json,
    deserialize=_interpolation_type_from_json,
)


def _data_type_to_json(data_type: Optional[KukurDataType]) -> Optional[str]:
    if data_type is None:
        return None
    return data_type.value


def _data_type_from_json(data_type: Optional[str]) -> Optional[KukurDataType]:
    if data_type is None:
        return None
    return KukurDataType(data_type)


DataType = MetadataField[Optional[KukurDataType]](
    "data type",
    default=None,
    serialized_name="dataType",
    serialize=_data_type_to_json,
    deserialize=_data_type_from_json,
)


DictionaryName = MetadataField[Optional[str]](
    "dictionary name", default=None, serialized_name="dictionaryName"
)


def _dictionary_to_json(
    dictionary: Optional[KukurDictionary],
) -> Optional[List[Tuple[int, str]]]:
    if dictionary is None:
        return None
    return list(dictionary.mapping.items())


def _dictionary_from_json(
    dictionary: Optional[List[Tuple[int, str]]]
) -> Optional[KukurDictionary]:
    if dictionary is None:
        return None
    return KukurDictionary(dict(dictionary))


Dictionary = MetadataField[Optional[KukurDictionary]](
    "dictionary",
    default=None,
    serialized_name="dictionary",
    serialize=_dictionary_to_json,
    deserialize=_dictionary_from_json,
)


def register_default_fields(cls) -> None:
    """Register all common metadata fields to the Metadata class."""
    cls.register_field(Description)
    cls.register_field(Unit)
    cls.register_field(LimitLowPhysical)
    cls.register_field(LimitHighPhysical)
    cls.register_field(LimitLowFunctional)
    cls.register_field(LimitHighFunctional)
    cls.register_field(Accuracy)
    cls.register_field(AccuracyPercentage)
    cls.register_field(InterpolationType)
    cls.register_field(DataType)
    cls.register_field(DictionaryName)
    cls.register_field(Dictionary)
